<?php


namespace ContainerCheck;

use Concrete\Core\Asset\Asset;
use Concrete\Core\Entity\Page\Container;
use Concrete\Core\Package\PackageService;
use Concrete\Core\Page\Container\TemplateLocator;
use Concrete\Core\Page\Page;
use Concrete\Core\Support\Facade\Application;
use Concrete\Core\Asset\AssetList;
use Doctrine\ORM\EntityManagerInterface;
use Concrete\Core\Support\Facade\Config;
use Exception;
use Throwable;
use ValueError;

defined('C5_EXECUTE') or die('Access Denied.');

/*
 * Software by John Liddiard (aka JohntheFish)
 * www.c5magic.co.uk
 *
 * This package contains software copyright and proprietary to John Liddiard
 *
 */

class ContainerChecker
{
    protected $container;
    protected $raw_container_html;
    protected $looped;
    protected $conditional;

    public function __construct($container)
    {
        $this->container = $container;
        $this->app = Application::getFacadeApplication();
    }

    public function containsLoop()
    {
        return !empty($this->looped);
    }

    public function containsCondition()
    {
        return !empty($this->conditional);
    }

    /*
     * HTML without area names
     */
    public function getContainerHTML()
    {
        return $this->cleanup_container_html($this->getContainerRAWHTML());
    }


    /*
     * Strip it down to only what is needed to represent the container
     */
    protected function cleanup_container_html($container_html)
    {
        /*
         * Nothing before the first <div>
         */
        $container_html = preg_replace("/^.*?<div/si", '<div', $container_html);

        /*
         * No php tags, though htmlawed should also catch such
         */
        $container_html = preg_replace("/<\?.*?\?>/si", '', $container_html);

        /*
         * Then a further cleanup with htmlawed
         */
        $limitations = [
            'safe'           => 1, // no scripts or handlers
            'tidy'           => '4s', // format, 4 spaces indentation
            'elements'       => 'div', // only <div> elements
            'deny_attribute' => '* -class', // only class attributes
        ];

        $container_html = htmLawed($container_html, $limitations);

        return $container_html;
    }

    /*
    * HTML with area names
    */
    public function getContainerHTMLWithAreas($edit=true)
    {
        return $this->cleanup_container_html_keeping_areas($this->getContainerRAWHTML(),$edit);
    }

    /*
     * This is the tricky variant because we want to keep area names, but it has to cope with
     * code tricks such as collapsable areas
     */
    protected function cleanup_container_html_keeping_areas($container_html, $edit=true)
    {
        $tokens = @token_get_all($container_html);

        $special = [];
        if (!empty($tokens)) {
            foreach ($tokens as $token) {
                if (is_array($token)) {
                    if ($token[0] == T_FOR) {
                        $special['FOR'] = 'FOR';
                    } elseif ($token[0] == T_FOREACH) {
                        $special['FOREACH'] = 'FOREACH';
                    } elseif ($token[0] == T_IF) {
                        $special['IF'] = 'IF';
                    }
                }
            }
        }
        if (!empty($special['FOR']) || !empty($special['FOREACH'])) {
            $this->looped = true;
        } else {
            $this->looped = false;
        }

        if (!empty($special['IF']) ||
            strpos($container_html, t('Optional')) !== false ||
            strpos($container_html, 'Optional') !== false ||
            strpos($container_html, '<code>') !== false ||
            strpos($container_html, '</code>') !== false
        ) {
            $this->conditional = true;
        } else {
            $this->conditional = false;
        }

        /*
         * Nothing before the first <div>
         */
        $container_html = preg_replace("/^.*?<div/si", '<div', $container_html);

        if (class_exists('\ContainerMagic\ContainerStyler')) {
            $cs = new \ContainerMagic\ContainerStyler($this->container);
            $container_html = $cs->parseAndFillContainerClasses($container_html, $edit); // no variant global classes at this level
        }

        $container_html = $this->parseAndFillContainerAreaNames($container_html);

        /*
         * Strip any remaining php tags, though htmlawed should also catch such
         */
        $container_html = preg_replace("/<\?.*?\?>/si", '', $container_html);

        /*
         * Then a further cleanup with htmlawed
         */
        $limitations = [
            'safe'           => 1, // no scripts or handlers
            'tidy'           => '4s', // format, 4 spaces indentation
            'elements'       => 'div,code', // only <div> elements
            'deny_attribute' => '* -class', // only class attributes
        ];

        $container_html = htmLawed($container_html, $limitations);
        if (!empty($this->conditional)) {
            $container_html = preg_replace("/<code>.*?<\/code>/", '<code>' . t('(Optional)') . '</code>', $container_html);
        }


        return $container_html;
    }

    /*
     * Find $variable used in any area name declarations, note the identifier, then replace ->display
     * of each such $variable with the area name. This needs to be non-greedy and only act within single
     * <? ... ?>
     */
    protected function parseAndFillContainerAreaNames($container_html)
    {

        if (preg_match_all("/<\?.*?\\\$([\w]+)\s*=.*?ContainerArea\(.*?\'([\w\s]+)\'.*?<\/div>/si", $container_html, $m)) {
            foreach ($m[1] as $ix => $area_var) {
                $area_name = $m[2][$ix];
                /*
                $sub = $m[0][$ix];
                if (preg_match("/<code>.*?'(.*)'.*?<\/code>/si", $sub, $mc)) {
                    $code_msg = $mc[1];
                    $area_name .= ' <code class="small">(' . $code_msg . ')</code>';
                }
                */

                $container_html = preg_replace("/<\?[^\?]*?" . $area_var . "\s*\-\>\s*display.*?\?>/si", $area_name, $container_html, 1);
            }
        }
        return $container_html;
    }

    public function getContainerExampleHTML()
    {
        $container_html = $this->cleanup_container_html_keeping_areas($this->getContainerRAWHTML(false));
        $limitations = [
            'safe'           => 1, // no scripts or handlers
            'tidy'           => '4s', // format, 4 spaces indentation
            'elements'       => 'div', // only <div> elements
            'deny_attribute' => '* -class', // only class attributes
        ];

        $container_html = htmLawed($container_html, $limitations);
        return $container_html;
    }


    public function getContainerHTMLWithAreasNoClasses()
    {
        return $this->cleanup_container_html_no_classes($this->getContainerRAWHTML(true));
    }

    protected function cleanup_container_html_no_classes($container_html)
    {
        $container_html = $this->cleanup_container_html_keeping_areas($container_html);
        $limitations = [
            'safe'           => 1, // no scripts or handlers
            'tidy'           => '4s', // format, 4 spaces indentation
            'elements'       => 'div', // only <div> elements
            'deny_attribute' => '*', // no attributes
        ];

        $container_html = htmLawed($container_html, $limitations);
        return $container_html;
    }

    /*
     * Locate a container's file, read it, clean up the html and return it.
     * Only read the file once, cache in a variable
     */
    public function getContainerRawHTML($edit=true)
    {
        if (!empty($this->raw_container_html)) {
            return $this->raw_container_html;
        }

        try {
            if (!empty($this->container)) {
                $locator = $this->app->make(TemplateLocator::class);
                $home = Page::getByID(Page::getHomePageID());
                $fh = $this->app->make('helper/file');
                $container_file = $locator->getFileToRender($home, $this->container);

                /*
                 * Integrate with Container Magic (if installed)
                 */
                if (class_exists('\ContainerMagic\ContainerFinder')) {
                    $override_theme = Config::get('jl_container_magic.override_theme');
                    if (empty($container_file) || $override_theme) {
                        $cf = app()->make('\ContainerMagic\ContainerFinder');
                        $container_file = $cf->getFileToRenderByTheme($this->container);
                    }
                }
            }

            if (!empty($container_file) && file_exists($container_file)) {
                $this->raw_container_html = trim($fh->getContents($container_file));
            }
        } catch (Exception|Throwable|ValueError $e) {
            // null
        }

        if (empty($this->raw_container_html)) {
            if (!empty($this->container) && is_object($this->container)) {
                return '<div class="alert alert-error">' . t('Unable to load container "%s"', $this->container->getContainerDisplayName()) . '</div>';
            } else {
                return '<div class="alert alert-error">' . t('Unable to load container') . '</div>';
            }

        }

        return $this->raw_container_html;
    }


    /*
     * Assets specific to container checker
     * Usage:
     *  ContainerChecker::register_assets();
     *  $this->requireAsset('container_check');
     */
    static public function register_assets()
    {
        $pkg = Application::getFacadeApplication()->make(PackageService::class)->getByHandle('jl_container_check');
        $al = AssetList::getInstance();

        $al->register('css',
                      'container_check',
                      'css/container_check.css',
                      [
                          'version'  => $pkg->getPackageVersion(),
                          'position' => Asset::ASSET_POSITION_HEADER,
                          'minify'   => true,
                          'combine'  => true,
                      ],
                      $pkg
        );

        $al->register('javascript',
                      'container_check',
                      'js/container_check.js',
                      [
                          'version'  => $pkg->getPackageVersion(),
                          'position' => Asset::ASSET_POSITION_FOOTER,
                          'minify'   => true,
                          'combine'  => true,
                      ],
                      $pkg
        );

        $al->registerGroup('container_check', [
            ['css', 'container_check'],
            ['javascript', 'jquery'],
            ['javascript', 'container_check'],
        ]);
    }

    /*
     * Purloined from the dashboard container list page
     */
    static public function list_containers()
    {
        if (class_exists('ContainerMagic\ContainerStyler')) {
            return \ContainerMagic\ContainerStyler::listContainers();
        }

        $entityManager = app()->make(EntityManagerInterface::class);
        $containers = $entityManager->getRepository(Container::class)->findBy([], ['containerName' => 'asc']);

        usort($containers, function ($a, $b) {
            /*
             * Sort needs to be by the clean name
             */
            $ah = preg_replace("/[^a-z0-9]+/i", ' ', strtolower($a->getContainerDisplayName()));
            $bh = preg_replace("/[^a-z0-9]+/i", ' ', strtolower($b->getContainerDisplayName()));
            if ($ah < $bh) {
                return -1;
            }
            if ($ah > $bh) {
                return 1;
            }
        });
        return $containers;
    }


}